@@ -5,16 +5,13 @@ use macro_rules_attribute::apply;
 use pyo3::exceptions::PyRuntimeError;
 use rr_util::{
     python_error_exception,
-    util::{is_unique, NamedAxes},
+    util::{is_unique, HashBytes, NamedAxes},
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 use thiserror::Error;
 use uuid::Uuid;
 
-use crate::{
-    circuit_utils::deep_map_pass_down_branching, deep_map_unwrap, prelude::*,
-    visit_circuit_with_parents_fallible,
-};
+use crate::{circuit_utils::deep_map_pass_down_branching, deep_map_unwrap, prelude::*, Module};
 
 pub fn set_named_axes<T: CircuitNode>(node: T, named_axes: NamedAxes) -> Result<T> {
     for name in named_axes.values() {
@@ -36,9 +33,10 @@ use pyo3::prelude::*;
 pub fn set_named_axes_py(circuit: CircuitRc, named_axes: NamedAxes) -> Result<CircuitRc> {
     set_named_axes(circuit, named_axes)
 }
+
 pub fn merge_named_axes(a: &NamedAxes, b: &NamedAxes) -> NamedAxes {
     let mut result = a.clone();
-    result.extend(b.clone().into_iter());
+    result.extend(b.clone());
     result
 }
 
@@ -46,8 +44,8 @@ pub fn named_axes_backward<T: CircuitNode + Clone>(
     circuit: &T,
     named_axes: &NamedAxes,
 ) -> Vec<NamedAxes> {
-    let child_axis_map = circuit.child_axis_map();
-    child_axis_map
+    circuit
+        .child_axis_map()
         .iter()
         .map(|z| {
             z.iter()
@@ -84,7 +82,7 @@ pub fn propagate_named_axes(
                     .map(|_c| Default::default())
                     .collect::<Vec<NamedAxes>>()
             } else {
-                named_axes_backward(&**circuit, &{ axes.clone() })
+                named_axes_backward(&**circuit, axes)
             };
             child_axis_names
         },
@@ -114,7 +112,6 @@ pub fn axis_of_name(circuit: &Circuit, name: &str) -> Option<usize> {
 pub fn get_axis_leaves(
     circuit: CircuitRc,
     axis: usize,
-    _error_on_removable_axis_leaf: bool,
 ) -> Result<(
     String,
     CircuitRc,
@@ -142,33 +139,65 @@ pub fn get_axis_leaves(
     }
     let circ_named_axes =
         propagate_named_axes(circuit, BTreeMap::from([(axis as u8, name.clone())]), true);
+
     let mut result_axis: HashMap<CircuitRc, usize> = Default::default();
     let mut result_no_axis: HashSet<CircuitRc> = Default::default();
-    visit_circuit_with_parents_fallible(circ_named_axes.clone(), |x, parents| {
-        // println!("child");
-        // x.print();
-        // println!("parents");
-        // for parent in parents {
-        //     parent.print();
-        // }
-        if let Some(i) = axis_of_name(&x,&name) && !x.children().any(|child|axis_of_name(&child, &name).is_some())
-        {
-            // if  error_on_removable_axis_leaf && get_removable_axes(&x.crc()).contains(&i){
-            //     return Err(CircuitConstructionError::ExpandingRemovableAxisUnfortunateError {  })
-            // }
-            result_axis.insert(x.clone(), i);
-        }
+    let mut seen: HashMap<HashBytes, CircuitRc> = Default::default();
 
-        if axis_of_name(&x, &name).is_none()
-            && parents
-                .iter()
-                .any(|child| axis_of_name(child, &name).is_some())
-        {
-            result_no_axis.insert(x);
+    fn f(
+        ra: &mut HashMap<CircuitRc, usize>,
+        rna: &mut HashSet<CircuitRc>,
+        seen: &mut HashMap<HashBytes, CircuitRc>,
+        name: &str,
+        x: CircuitRc,
+    ) -> Result<CircuitRc> {
+        let h = x.info().hash;
+        if let Some(r) = seen.get(&h) {
+            Ok(r.clone())
+        } else {
+            {
+            if axis_of_name(&x, name).is_none() {
+                rna.insert(x.clone());
+                return Ok(x);
+            }
+
+            if let Circuit::Module(m) = &**x {
+                if axis_of_name(&x, name).unwrap() < m.aligned_batch_shape().len() {
+                    let nodes2 = m
+                        .nodes
+                        .iter()
+                        .map(|c| f(ra, rna, seen, name, c.clone()))
+                        .collect::<Result<Vec<CircuitRc>>>()?;
+                    return Ok(Module::nrc(nodes2, m.spec.clone(), m.name_cloned()));
+                } else {
+                    bail!("batch_to_concat/get_axis_leaves: modules not fully supported, please substitute first")
+                }
+            }
+
+            if !x
+                .children()
+                .any(|child| axis_of_name(&child, name).is_some())
+            {
+                if let Some(i) = axis_of_name(&x, name) {
+                    ra.insert(x.clone(), i);
+                    return Ok(x);
+                } else {
+                    panic!("get_axis_leaves: axis gone before traversal done")
+                }
+            }
+
+            x.map_children(|c| f(ra, rna, seen, name, c))
+        }.map(|r2|{seen.insert(h, r2.clone()); r2})
         }
-        Ok(())
-    })?;
-    Ok((name, circ_named_axes, result_axis, result_no_axis))
+    }
+    let circ_new = f(
+        &mut result_axis,
+        &mut result_no_axis,
+        &mut seen,
+        &name,
+        circ_named_axes.clone(),
+    )?;
+    Ok((name, circ_new, result_axis, result_no_axis))
 }
 
 /// remove all instances of this axis name from circuit
