@@ -1,12 +1,13 @@
 use std::{
-    fmt::Debug,
+    fmt::{Debug, Write},
     hash::{Hash, Hasher},
 };
 
 use anyhow::{anyhow, bail, Context, Result};
+use itertools::Itertools;
 use pyo3::prelude::*;
 use rand::Rng;
-use rr_util::{fn_struct, py_types::MaybeNotSet, simple_default, util::hash_to_string};
+use rr_util::{fn_struct, py_types::MaybeNotSet, simple_default};
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 
 use crate::{
@@ -53,8 +54,6 @@ pub struct PrintHtmlOptions {
     #[pyo3(get, set)]
     pub primary_color: String,
     #[pyo3(get, set)]
-    pub use_serial_numbers: bool,
-    #[pyo3(get, set)]
     pub number_leaves: bool,
     #[pyo3(get, set)]
     pub keep_all_cells: bool,
@@ -73,7 +72,6 @@ impl Default for PrintHtmlOptions {
             display_copy_button: true,
             colorer: PrintHtmlOptions::type_colorer(),
             primary_color: "lightgray".to_owned(),
-            use_serial_numbers: false,
             number_leaves: false,
             keep_all_cells: true,
             overall_traversal: Some(OpaqueIterativeMatcherVal::never()),
@@ -87,43 +85,61 @@ impl PrintHtmlOptions {
     pub fn fill_overall_data(
         &self,
         circuits: Vec<CircuitRc>,
-        data: &mut HashMap<HashBytes, NodeJsonData>,
+        nodes_string: &mut String,
+        seen: &mut HashSet<HashBytes>,
     ) -> Result<()> {
         fn recurse(
             circ: CircuitRc,
-            data: &mut HashMap<HashBytes, NodeJsonData>,
+            nodes_string: &mut String,
+            seen: &mut HashSet<HashBytes>,
             selfy: &PrintHtmlOptions,
-            traversal: Option<OpaqueIterativeMatcherVal>,
+            overall_traversal: Option<OpaqueIterativeMatcherVal>,
             node_comments: Vec<String>,
+            parent_info: Option<(&Circuit, usize)>,
         ) -> Result<()> {
-            if let Some(_) = data.get(&circ.info().hash) {
+            let printer = PrintOptions::default();
+            if !seen.insert(circ.info().hash) {
                 return Ok(());
             }
-            let mut desc = selfy.repr_line(circ.clone())?;
+            let mut desc: String = selfy.repr_line(circ.clone())?;
+            parent_info.map(|(parent, child_idx)| {
+                // get_child_info is very tied to regular printer, hand rolling
+                parent.as_module().map(|m| {
+                    child_idx.checked_sub(1).map(|i| {
+                        let arg_spec = &m.spec.arg_specs[i];
+                        arg_spec.symbol.name().map(|n| write!(desc, " ! '{}'", n));
+                    })
+                });
+            });
             write_comment(&mut desc, node_comments, &selfy.commenters, circ.crc())?;
-            data.insert(
-                circ.info().hash,
-                NodeJsonData {
-                    name: circ.name_cloned().unwrap_or("".to_owned()),
-                    desc,
-                    color: match selfy
-                        .colorer
-                        .call(circ.clone())
-                        .context("colorer errored in html printer")?
-                    {
-                        Some(col) => col,
-                        None => selfy.primary_color.clone(),
-                    },
-                    children: circ
-                        .clone()
-                        .children()
-                        .map(|c| hash_to_string(&c.info().hash))
-                        .collect(),
-                },
-            );
+            make_str_safe(&mut desc)?;
+            let mut color = match selfy
+                .colorer
+                .call(circ.clone())
+                .context("colorer errored in html printer")?
+            {
+                Some(col) => col,
+                None => selfy.primary_color.clone(),
+            };
+            make_str_css_safe(&mut color)?;
+            let mut name = circ.name_cloned().unwrap_or("".to_owned());
+            make_str_safe(&mut name)?;
+            write!(
+                nodes_string,
+                "\"{}\":{{name:\"{}\",desc:\"{}\",color:\"{}\",children:[",
+                &circ.hash_base16(),
+                name,
+                desc,
+                &color,
+            )
+            .unwrap();
+            for c in circ.children() {
+                write!(nodes_string, "\"{}\",", c.hash_base16()).unwrap();
+            }
+            nodes_string.push_str("]},");
 
             let new_traversal_per = if let Some(new_traversal_per) =
-                op_traversal_per(traversal, circ.crc())
+                op_traversal_per(overall_traversal, circ.crc())
                     .context("traversal errored in html printer")?
             {
                 new_traversal_per
@@ -131,91 +147,91 @@ impl PrintHtmlOptions {
                 // all finished case
                 return Ok(());
             };
-
+            let give_parent_info = printer.has_child_info(&circ);
             for (i, (child, new_traversal)) in circ.children().zip(new_traversal_per).enumerate() {
                 let child_specific_commenters = if selfy.comment_arg_names {
                     get_child_comments(&(**circ), i)
                 } else {
                     vec![]
                 };
-                recurse(child, data, selfy, new_traversal, child_specific_commenters)?;
+                recurse(
+                    child,
+                    nodes_string,
+                    seen,
+                    selfy,
+                    new_traversal,
+                    child_specific_commenters,
+                    if give_parent_info {
+                        Some((&circ, i))
+                    } else {
+                        None
+                    },
+                )?;
             }
             Ok(())
         }
         for circuit in circuits.clone() {
-            recurse(circuit, data, &self, self.overall_traversal.clone(), vec![])?;
+            recurse(
+                circuit,
+                nodes_string,
+                seen,
+                &self,
+                self.overall_traversal.clone(),
+                vec![],
+                None,
+            )?;
         }
         Ok(())
     }
     pub fn fill_to_expand_data(
         &self,
         circuits: Vec<CircuitRc>,
-        nodes_to_expand: &mut Vec<HashBytes>, // Nodes you can see the children of
-        data: &mut HashMap<HashBytes, NodeJsonData>, /* Data of all nodes which you might want to see */
+        to_expand_string: &mut String, // Nodes you can see the children of
+        seen: &HashSet<HashBytes>,     /* Data of all nodes which you might want to see */
     ) -> Result<()> {
         let mut expanded_nodes: HashSet<HashBytes> = Default::default(); // Nodes that are seen from the beginning
         fn recurse(
             circ: CircuitRc,
-            nodes_to_expand: &mut Vec<HashBytes>,
+            to_expand_string: &mut String, // Nodes you can see the children of
             expanded_nodes: &mut HashSet<HashBytes>,
-            data: &mut HashMap<HashBytes, NodeJsonData>,
-            runnning_position_number: &mut usize,
             selfy: &PrintHtmlOptions,
             traversal: Option<OpaqueIterativeMatcherVal>,
+            seen: &HashSet<HashBytes>,
         ) -> Result<()> {
-            if let Some(_) = expanded_nodes.get(&circ.info().hash) {
+            let key = circ.info().hash;
+            if !expanded_nodes.insert(key) {
                 return Ok(());
             }
-            expanded_nodes.insert(circ.info().hash);
-
             let new_traversal_per_op = op_traversal_per(traversal, circ.crc())?;
 
-            let mut desc = String::new();
-            if selfy.use_serial_numbers {
-                desc.push_str(&runnning_position_number.to_string());
-                desc.push(' ');
-                *runnning_position_number += 1;
-            }
-            if let Some(old_data) = data.get_mut(&circ.info().hash) {
-                desc.push_str(&old_data.desc);
-                old_data.desc = desc; // Overwrite description to have the right nodes_to_expand count
-            } else {
-                bail!(anyhow!("overall traversal does not cover traversal."))
-            }
-
             let new_traversal_per = if let Some(new_traversal_per) = new_traversal_per_op {
                 new_traversal_per
             } else {
                 // all finished case
                 return Ok(());
             };
-            // below is quadratic time due to contains on vec
-            if !nodes_to_expand.contains(&circ.info().hash) {
-                nodes_to_expand.push(circ.info().hash);
-            }
+            write!(to_expand_string, "\"{}\",", &circ.hash_base16()).unwrap();
+
             for (child, new_traversal) in circ.children().zip(new_traversal_per) {
                 recurse(
                     child,
-                    nodes_to_expand,
+                    to_expand_string,
                     expanded_nodes,
-                    data,
-                    runnning_position_number,
                     selfy,
                     new_traversal.clone(),
+                    seen,
                 )?;
             }
             Ok(())
         }
-        let mut runnning_position_number = 0;
         for circuit in circuits {
             recurse(
                 circuit,
-                nodes_to_expand,
+                to_expand_string,
                 &mut expanded_nodes,
-                data,
-                &mut runnning_position_number,
                 &self,
                 self.traversal.clone(),
+                seen,
             )?;
         }
         Ok(())
@@ -239,7 +255,6 @@ impl PrintHtmlOptions {
         display_copy_button = "PrintHtmlOptions::default().display_copy_button",
         colorer = "PrintHtmlOptions::default().colorer",
         primary_color = "PrintHtmlOptions::default().primary_color",
-        use_serial_numbers = "PrintHtmlOptions::default().use_serial_numbers",
         number_leaves = "PrintHtmlOptions::default().number_leaves",
         keep_all_cells = "PrintHtmlOptions::default().keep_all_cells",
         overall_traversal = "PrintHtmlOptions::default().overall_traversal.unwrap()",
@@ -252,7 +267,6 @@ impl PrintHtmlOptions {
         display_copy_button: bool,
         colorer: CircuitToColor,
         primary_color: String,
-        use_serial_numbers: bool,
         number_leaves: bool,
         keep_all_cells: bool,
         overall_traversal: Option<OpaqueIterativeMatcherVal>,
@@ -265,7 +279,6 @@ impl PrintHtmlOptions {
             display_copy_button,
             colorer,
             primary_color,
-            use_serial_numbers,
             number_leaves,
             keep_all_cells,
             overall_traversal,
@@ -281,7 +294,6 @@ impl PrintHtmlOptions {
         colorer = "MaybeNotSet(None)",
         primary_color = "Default::default()",
         display_copy_button = "Default::default()",
-        use_serial_numbers = "Default::default()",
         number_leaves = "Default::default()",
         keep_all_cells = "Default::default()",
         overall_traversal = "Default::default()",
@@ -295,7 +307,6 @@ impl PrintHtmlOptions {
         display_copy_button: MaybeNotSet<bool>,
         colorer: MaybeNotSet<CircuitToColor>,
         primary_color: MaybeNotSet<String>,
-        use_serial_numbers: MaybeNotSet<bool>,
         number_leaves: MaybeNotSet<bool>,
         keep_all_cells: MaybeNotSet<bool>,
         overall_traversal: MaybeNotSet<Option<OpaqueIterativeMatcherVal>>,
@@ -311,7 +322,6 @@ impl PrintHtmlOptions {
             display_copy_button: display_copy_button.0.unwrap_or(cloned.display_copy_button),
             colorer: colorer.0.unwrap_or(cloned.colorer),
             primary_color: primary_color.0.unwrap_or(cloned.primary_color),
-            use_serial_numbers: use_serial_numbers.0.unwrap_or(cloned.use_serial_numbers),
             number_leaves: number_leaves.0.unwrap_or(cloned.number_leaves),
             keep_all_cells: keep_all_cells.0.unwrap_or(cloned.keep_all_cells),
             overall_traversal: overall_traversal.0.unwrap_or(cloned.overall_traversal),
@@ -366,23 +376,9 @@ impl PrintHtmlOptions {
 
     #[args(circuits = "*")]
     pub fn repr(&self, circuits: Vec<CircuitRc>) -> Result<String> {
-        let mut nodes_to_expand: Vec<HashBytes> = Default::default();
-        let mut data: HashMap<HashBytes, NodeJsonData> = Default::default();
-        self.fill_overall_data(circuits.clone(), &mut data)?;
-        self.fill_to_expand_data(circuits.clone(), &mut nodes_to_expand, &mut data)?;
-
         let mut html = HTML_HEADER.to_owned();
-
-        // var toExpandtoggle___js___ = ['1', ...];
-        html.push_str("var toExpandtoggle___js___ = [");
-        html.push_str(
-            &nodes_to_expand
-                .iter()
-                .map(|h| format!("\"{}\"", hash_to_string(h)))
-                .collect::<Vec<String>>()
-                .join(","),
-        );
-        html.push_str("];");
+        let mut primary_color = self.primary_color.clone();
+        make_str_css_safe(&mut primary_color)?;
 
         // var data = {1: {
         //     name: 'I k2 logits_diff, true I',
@@ -391,22 +387,21 @@ impl PrintHtmlOptions {
         //     children: ['2', '3'],
         //   }
         html.push_str("var datatoggle___js___ = {");
-        html.push_str(
-            &data
-                .iter()
-                .map(|(h, json_data)| Ok(hash_to_string(h) + ":" + &json_data.to_string()?))
-                .collect::<Result<Vec<String>>>()?
-                .join(","),
-        );
+        let mut seen: HashSet<HashBytes> = HashSet::default();
+        self.fill_overall_data(circuits.clone(), &mut html, &mut seen)?;
         html.push_str("};");
 
+        // var toExpandtoggle___js___ = ['1', ...];
+        html.push_str("var toExpandtoggle___js___ = [");
+        self.fill_to_expand_data(circuits.clone(), &mut html, &seen)?;
+        html.push_str("];");
+
         // var rootstoggle___js___ = ['1', ...];
         html.push_str("var rootstoggle___js___ = [");
         html.push_str(
             &circuits
                 .iter()
-                .map(|c| format!("\"{}\"", hash_to_string(&c.info().hash)))
-                .collect::<Vec<String>>()
+                .map(|c| format!("\"{}\"", &c.hash_base16()))
                 .join(","),
         );
         html.push_str("];");
@@ -430,14 +425,11 @@ impl PrintHtmlOptions {
 
         html.push_str(CSS_HEADER);
 
-        html.push_str(&format!(
-            "--primary: {};",
-            make_str_css_safe(&self.primary_color)?
-        ));
+        html.push_str(&format!("--primary: {};", primary_color,));
         html.push_str(CSS);
 
         html.push_str(&format!(
-            "<div class=\"mainUL\" id=\"s{seed}\"><ul></ul></div>",
+            "<div class=\"mainDiv\" id=\"s{seed}\"><ul class=\"list\"></ul></div>",
         ));
         html.push_str(JS);
 
@@ -473,7 +465,6 @@ impl PrintHtmlOptions {
             MaybeNotSet(None),
             MaybeNotSet(None),
             MaybeNotSet(None),
-            MaybeNotSet(None),
         )
     }
 
@@ -574,48 +565,29 @@ fn color_from_colorcode(code: usize) -> String {
     return COLORCODE_TO_RGB[code].to_owned();
 }
 
-pub struct NodeJsonData {
-    name: String,
-    desc: String,
-    color: String,
-    children: Vec<String>,
-}
-
-fn make_str_safe(s: &String) -> Result<String> {
+// this does mutation to avoid copying in the 99% case when it doesn't change anything
+fn make_str_safe(s: &mut String) -> Result<()> {
     if s.contains("___js___") {
         bail!(anyhow!("___js___ is a reserved name. Don't use it."))
     }
     if s.contains("\"") {
         bail!(anyhow!("Can't use quotes in string printed to html."))
     }
-    Ok(s.replace("<", "&lt").replace(">", "&gt"))
-}
-fn make_str_css_safe(s: &String) -> Result<String> {
-    return Ok(make_str_safe(s)
-        .context("wrong color format (no quotes allowed)")?
-        .replace(";", ""));
-}
-
-impl NodeJsonData {
-    fn to_string(&self) -> Result<String> {
-        let mut r = String::from("{name:\"");
-        r.push_str(&make_str_safe(&self.name)?);
-        r.push_str("\", desc:\"");
-        r.push_str(&make_str_safe(&self.desc)?);
-        r.push_str("\", color:\"");
-        r.push_str(&make_str_css_safe(&self.color)?);
-        r.push_str("\", children:[");
-        r.push_str(
-            &self
-                .children
-                .iter()
-                .map(|c| format!("\"{c}\""))
-                .collect::<Vec<String>>()
-                .join(","),
-        );
-        r.push_str("]}");
-        Ok(r)
+    if s.contains("<") || s.contains(">") {
+        let newstring = s.replace("<", "&lt").replace(">", "&gt");
+        s.truncate(0);
+        s.push_str(&newstring);
     }
+    Ok(())
+}
+fn make_str_css_safe(s: &mut String) -> Result<()> {
+    make_str_safe(s).context("wrong color format (no quotes allowed)")?;
+    if s.contains(";") {
+        let newstring = s.replace(";", "");
+        s.truncate(0);
+        s.push_str(&newstring);
+    }
+    return Ok(());
 }
 
 // Variables that shouldn't be overwritten between cells should have ___js___ as suffix.
@@ -628,22 +600,18 @@ const CSS_HEADER: &str = r#"
     :root {"#;
 const CSS: &str = r#"
     }
-    ul,
-    .mainUL {
+    .mainDiv {
+      margin: 0;
+      padding: 0;
+    }
+    .list.list {
+      margin: 0;
+      padding: 0;
       list-style-type: none;
-    }
-    .mainUL {
-      margin: 0;
-      padding: 0;
-    }
-    ul,
-    li {
-      margin: 0;
-      padding: 0;
       color: var(--primary);
     }
-    .caret,
-    .caret-end {
+    .expand-caret.expand-caret.expand-caret,
+    .expand-caret-end.expand-caret-end.expand-caret-end {
       cursor: pointer;
       -webkit-user-select: none; /* Safari 3.1+ */
       -moz-user-select: none; /* Firefox 2+ */
@@ -652,36 +620,36 @@ const CSS: &str = r#"
       margin-left: 0.2em;
       margin-right: 0.4em;
     }
-    .caret::before,
-    .caret-end::before {
+    .expand-caret::before,
+    .expand-caret-end::before {
       content: '>';
       color: var(--primary);
       display: inline-block;
     }
-    .caret-end::before {
+    .expand-caret-end::before {
       color: var(--primary);
       opacity: 0.5;
     }
-    .caret-down::before {
+    .expand-caret-down::before {
       -ms-transform: rotate(90deg);
       -webkit-transform: rotate(90deg);
       transform: rotate(90deg);
       color: var(--primary);
     }
-    .nested {
+    .nested.nested.nested {
       display: none;
     }
-    .active {
+    .active.active.active {
       padding-left: 0.2em;
       margin-left: 0.42em;
 
       display: block;
     }
-    .active.onlychild {
+    .active.active.onlychild {
       margin-left: 0em;
       padding-left: 0em;
     }
-    button {
+    .copy-button {
       margin: 0px 0.2em 0px 0.2em;
       padding: 0px;
       background-color: transparent;
@@ -690,16 +658,16 @@ const CSS: &str = r#"
       outline: none;
       opacity: 0.6;
     }
-    button:hover {
+    .copy-button:hover {
       opacity: 1;
     }
-    button:active {
+    .copy-button:active {
       background-color: var(--primary);
     }
-    button:focus {
+    .copy-button:focus {
       outline: none;
     }
-    pre {
+    .inline-pre {
       margin: 0;
       padding: 0;
       display: inline;
@@ -721,13 +689,14 @@ const JS: &str = r#"
   // Parameters: id of the parent html element, id of the node to expand
   function addChildtoggle___js___(parent_id, id) {
     var child = document.createElement('li');
+    child.classList.add('list');
 
     // Example of the generated inner HTML:
-    // <span class="caret" onclick="toggle('1')"></span>
-    // <pre style="color: hsl(70, 100%, 70%)">1 'I k2 logits_diff, true I' [t: Cumulant] [s: ()]</pre>
-    // <pre style="color: magenta"></pre>
-    // <button onclick="toClipboard('I k2 logits_diff, true I')">📋</button>
-    // <ul class="nested" style="border-left: 3px solid hsl(70, 100%, 70%)"></ul>;
+    // <span class="expand-caret" onclick="toggle('1')"></span>
+    // <pre class="inline-pre" style="color: hsl(70, 100%, 70%)">1 'I k2 logits_diff, true I' [t: Cumulant] [s: ()]</pre>
+    // <pre class="inline-pre" style="color: magenta"></pre>
+    // <button onclick="toClipboard('I k2 logits_diff, true I')" class="copy-button">📋</button>
+    // <ul class="nested list" style="border-left: 3px solid hsl(70, 100%, 70%)"></ul>;
 
     const childData = datatoggle___js___[id];
     if (!childData.displayedCount) {
@@ -741,26 +710,26 @@ const JS: &str = r#"
       desc += ' (repeat)';
     }
     var copyButton = displayCopyButtontoggle___js___
-      ? `<button onclick="toClipboard('` + childData.name + `')">📋</button>`
+      ? `<button onclick="toClipboard('` + childData.name + `')" class="copy-button">📋</button>`
       : '';
 
     var onlyChild = datatoggle___js___[id].children.length == 0;
     var expandForbidden = !expandClonestoggle___js___ && childData.displayedCount > 0;
     var cantExpand = onlyChild || expandForbidden;
-    var caretClass = cantExpand ? 'caret-end' : 'caret';
+    var caretClass = cantExpand ? 'expand-caret-end' : 'expand-caret';
 
     var innerHTML =
       `<span class="` +
       caretClass +
       `" onclick="toggletoggle___js___('` +
       elementId +
-      `')"></span><pre style="color: ` +
+      `')"></span><pre class="inline-pre" style="color: ` +
       childData.color +
       `">` +
       desc +
-      `</pre><pre style="color:magenta"></pre>` +
+      `</pre><pre class="inline-pre" style="color:magenta"></pre>` +
       copyButton +
-      `<ul class="nested" style="border-left: 3px solid ` +
+      `<ul class="nested list" style="border-left: 3px solid ` +
       childData.color +
       `"></ul>`;
     child.innerHTML = innerHTML;
@@ -786,7 +755,7 @@ const JS: &str = r#"
     var expandAllowed = expandClonestoggle___js___ || elementId.split('@')[1] == '0';
     if (!noChild && expandAllowed) {
       getLastChild(elt).classList.toggle('active');
-      elt.children[0].classList.toggle('caret-down');
+      elt.children[0].classList.toggle('expand-caret-down');
 
       let alreadExpanded = getLastChild(elt).children.length > 0;
       if (!alreadExpanded) {
