@@ -41,24 +41,21 @@ pub struct Tag {
 impl Tag {
     #[apply(new_rc)]
     pub fn new(node: CircuitRc, uuid: Uuid, name: Option<String>) -> (Self) {
-        let mut out = Self {
+        let out = Self {
             node,
             uuid,
-            name: Default::default(),
+            name,
             info: Default::default(),
         };
-        out.name = out.auto_name(name);
-        out.init_info().unwrap()
+        out.initial_init_info().unwrap()
     }
 }
 
 impl CircuitNodeAutoName for Tag {
     const PRIORITY: OperatorPriority = OperatorPriority::PostFix {};
 
-    fn auto_name(&self, name: Option<String>) -> Option<String> {
-        name.or_else(|| {
-            child_name_with_maybe_paren(&Self::PRIORITY, self.node.clone()).map(|n| n + " tag")
-        })
+    fn auto_name(&self) -> Option<String> {
+        child_name_with_maybe_paren(&Self::PRIORITY, self.node.clone()).map(|n| n + " tag")
     }
 }
 
@@ -156,14 +153,13 @@ impl DiscreteVar {
                 probs: probs_and_group.info().shape[0],
             });
         }
-        let mut out = Self {
+        let out = Self {
             values,
             probs_and_group,
-            name: Default::default(),
+            name,
             info: Default::default(),
         };
-        out.name = out.auto_name(name);
-        out.init_info()
+        out.initial_init_info()
     }
 
     pub fn new_for_py(
@@ -183,11 +179,9 @@ impl DiscreteVar {
 impl CircuitNodeAutoName for DiscreteVar {
     const PRIORITY: OperatorPriority = OperatorPriority::PostFix {};
 
-    fn auto_name(&self, name: Option<String>) -> Option<String> {
+    fn auto_name(&self) -> Option<String> {
         // Ignores probs and groups because we almost never care about the probs and groups name
-        name.or_else(|| {
-            child_name_with_maybe_paren(&Self::PRIORITY, self.values.clone()).map(|n| n + " var")
-        })
+        child_name_with_maybe_paren(&Self::PRIORITY, self.values.clone()).map(|n| n + " var")
     }
 }
 
@@ -199,7 +193,9 @@ impl CircuitNodeComputeInfoImpl for DiscreteVar {
     }
 
     fn compute_flags(&self) -> CircuitFlags {
-        CircuitFlags::CAN_BE_SAMPLED
+        (self.compute_flags_default() | CircuitFlags::CAN_BE_SAMPLED)
+            & !CircuitFlags::IS_CONSTANT
+            & !CircuitFlags::IS_EXPLICITLY_COMPUTABLE
     }
 }
 
@@ -307,14 +303,13 @@ impl StoredCumulantVar {
                 },);
             }
         }
-        let mut out = Self {
+        let out = Self {
             cumulants,
             uuid,
-            name: Default::default(),
+            name,
             info: Default::default(),
         };
-        out.name = out.auto_name(name);
-        out.init_info()
+        out.initial_init_info()
     }
 
     pub fn new_mv(
@@ -336,25 +331,23 @@ impl StoredCumulantVar {
 }
 
 impl CircuitNodeAutoName for StoredCumulantVar {
-    fn auto_name(&self, name: Option<String>) -> Option<String> {
-        name.or_else(|| {
-            if self.cumulants.values().any(|x| x.name().is_none()) {
-                None
-            } else {
-                Some(
-                    "StoredCumulantVar(".to_owned()
-                        + &self
-                            .cumulants
-                            .iter()
-                            .map(|(r, x)| {
-                                format!("{r}: {}", Self::shorten_child_name(x.name().unwrap()))
-                            })
-                            .collect::<Vec<String>>()
-                            .join(", ")
-                        + ")",
-                )
-            }
-        })
+    fn auto_name(&self) -> Option<String> {
+        if self.cumulants.values().any(|x| x.name().is_none()) {
+            None
+        } else {
+            Some(
+                "StoredCumulantVar(".to_owned()
+                    + &self
+                        .cumulants
+                        .iter()
+                        .map(|(r, x)| {
+                            format!("{r}: {}", Self::shorten_child_name(x.name().unwrap()))
+                        })
+                        .collect::<Vec<String>>()
+                        .join(", ")
+                    + ")",
+            )
+        }
     }
 }
 
@@ -366,10 +359,14 @@ impl CircuitNodeComputeInfoImpl for StoredCumulantVar {
     }
 
     fn compute_flags(&self) -> CircuitFlags {
+        let flags = self.compute_flags_default()
+            & !CircuitFlags::IS_CONSTANT
+            & !CircuitFlags::IS_EXPLICITLY_COMPUTABLE;
         if self.cumulants.keys().max().unwrap() <= &2 {
-            return CircuitFlags::CAN_BE_SAMPLED;
+            flags | CircuitFlags::CAN_BE_SAMPLED
+        } else {
+            flags & !CircuitFlags::CAN_BE_SAMPLED
         }
-        CircuitFlags::NONE
     }
 }
 
