@@ -1,9 +1,7 @@
-use anyhow::{Context, Result};
+use anyhow::Result;
 use circuit_base::{prelude::*, CircuitType, Index};
 use circuit_rewrites::algebraic_rewrite::{index_elim_identity, push_down_index_raw};
-use get_update_node::{
-    iterative_matcher::require_single, IterativeMatcher, IterativeMatcherRc, Matcher,
-};
+use get_update_node::{IterativeMatcher, IterativeMatcherRc, Matcher};
 use macro_rules_attribute::apply;
 use pyo3::prelude::*;
 use rr_util::{cached_lambda, util::HashBytes};
@@ -46,19 +44,19 @@ pub fn push_down_index(
             .match_iterate(index.node.clone())?
             .unwrap_or_same(traversal)
             .0;
-        if let Some(new_traversal) = require_single(updated)
-            .context("push down index doesn't support traversal per child")?
-            .0
-        {
-            push_down_index_raw(
-                &index,
-                allow_partial_pushdown,
-                |new_index| push_down_rec(new_index, new_traversal.clone()),
-                suffix.clone(),
-            )
-        } else {
-            Ok(index.rc())
+
+        if updated.all_finished() {
+            return Ok(index.rc());
         }
+
+        let traversal_per_child = updated.per_child_with_term(index.node.num_children());
+
+        push_down_index_raw(
+            &index,
+            allow_partial_pushdown,
+            |i, new_index| push_down_rec(new_index, traversal_per_child[i].clone()),
+            suffix.clone(),
+        )
     }
 
     push_down_rec(index, traversal)
