@@ -63,7 +63,7 @@ impl From<usize> for SymbolicSizeProduct {
             if running_value < *p {
                 break;
             }
-            if running_value % p == 0 {
+            while running_value % p == 0 {
                 symbolic_sizes.push(i);
                 running_value /= p;
             }
