@@ -32,14 +32,6 @@ impl IntoPy<PyObject> for HashBytesToPy {
     }
 }
 
-pub fn hash_to_string(hash: &HashBytes) -> String {
-    hash.iter()
-        .map(|b| {
-            alphabet()[((*b) % 52) as usize].clone() + &alphabet()[((*b) / 52) as usize].clone()
-        })
-        .join("")
-}
-
 /// letters used in einsum and rearrange strings
 /// i wanted this to be a constant, but that is annoying in rust bc can't do any computation to produce constants
 /// so instead it's cached function
@@ -610,24 +602,31 @@ where
 
 #[macro_export]
 macro_rules! fn_struct{
-    ($vis:vis $name:ident : Fn($($iname:ident:$ity:ty),*)->$retty:ty)=>{
+    ($vis:vis $name:ident : $fn_ty:tt($($iname:ident:$ity:ty),*)->$retty:ty
+     $(;
+    {
+        $(
+        $enum_name:ident ($type_name:ty),
+        )*
+    })?
+     )=>{
         paste::paste!{
             #[derive(Clone,Debug)]
             $vis enum $name{
-                Py($crate::py_types::PyCallable),
                 Dyn([< $name DynStruct >]),
+                Py($crate::py_types::PyCallable),
+                $($(
+                    $enum_name($type_name),
+                )*)*
             }
             impl $name{
-                fn call(&self,$($iname:$ity),*)->anyhow::Result<$retty>{
-                    match &self{
-                        $name::Py(pyobj)=>rr_util::pycall!(pyobj,($($iname,)*),anyhow),
-                        $name::Dyn(dynf)=>dynf.0($($iname),*),
-                    }
-                }
                 fn new_dyn(f:Box<dyn Fn($($ity),*)->anyhow::Result<$retty> + Send + Sync>)->Self{
                     Self::Dyn([< $name DynStruct >] (std::sync::Arc::new(f)))
                 }
             }
+
+            $crate::fn_struct!(@mk_call $name $fn_ty($($iname:$ity),*)->$retty; ($($($enum_name,)*)*));
+
             #[pyclass]
             #[derive(Clone)]
             $vis struct [< $name DynStruct >] (
@@ -642,14 +641,20 @@ macro_rules! fn_struct{
             }
             impl<'source> pyo3::FromPyObject<'source> for $name{
                 fn extract(inp: &'source pyo3::PyAny) -> pyo3::PyResult<Self> {
+                    if let Ok(d) = inp.extract() {
+                        return Ok(Self::Dyn(d));
+                    }
                     Ok(Self::Py(inp.extract()?))
                 }
             }
             impl IntoPy<PyObject> for $name{
                 fn into_py(self, py: Python<'_>) -> PyObject {
                     match self{
-                        $name::Py(x)=>x.into_py(py),
-                        $name::Dyn(dynf)=>dynf.into_py(py)
+                        Self::Py(x)=>x.into_py(py),
+                        Self::Dyn(dynf)=>dynf.into_py(py),
+                        $($(
+                        Self::$enum_name(x)=>x.into_py(py),
+                        )*)*
                     }
                 }
             }
@@ -659,7 +664,52 @@ macro_rules! fn_struct{
                 }
             }
         }
+    };
+    (@mk_call $name:ident Fn($($iname:ident:$ity:ty),*) -> $retty:ty; ($($enum_name:ident,)*)) => {
+        impl $name {
+            fn call(&self,$($iname:$ity),*)->anyhow::Result<$retty>{
+                // nested macro to deal with expansion resolution
+                #[allow(unused_macros)]
+                macro_rules! stuff {
+                    ($x:expr) => (
+                        $x.call($($iname,)*)
+                    )
+                }
+
+                match &self{
+                    Self::Py(pyobj)=>rr_util::pycall!(pyobj,($($iname,)*),anyhow),
+                    Self::Dyn(dynf)=>dynf.0($($iname),*),
+                    $(
+                        Self::$enum_name(x)=>stuff!(x),
+                    )*
+                }
+            }
+        }
+    };
+    (@mk_call $name:ident FnMut($($iname:ident:$ity:ty),*) -> $retty:ty; ($($enum_name:ident,)*)) => {
+        impl $name {
+            fn call(&mut self,$($iname:$ity),*)->anyhow::Result<$retty>{
+                // nested macro to deal with expansion resolution
+                #[allow(unused_macros)]
+                macro_rules! stuff {
+                    ($x:expr) => (
+                        $x.call($($iname,)*)
+                    )
+                }
+
+                match self{
+                    Self::Py(pyobj)=>rr_util::pycall!(pyobj,($($iname,)*),anyhow),
+                    Self::Dyn(dynf)=>dynf.0($($iname),*),
+                    $(
+                        Self::$enum_name(x)=>stuff!(x),
+                    )*
+                }
+            }
+        }
+
     }
+
+
 }
 
 pub fn indent(x: String, width: usize) -> String {
